module Diff
  ( Tok
  , Op
  , diffToks
  , hunks
  , pprDiff
  , pprDiff1
  , pprHunk
  , pprHunks
  ) where

import Control.Monad
import qualified Data.ByteString as B
import qualified Data.ByteString.Builder as BB
import qualified Data.ByteString.Internal as BI
import qualified Data.ByteString.Lazy as BL
import Data.ByteString.UTF8 (fromString)
import Data.Function (on)
import Data.List (groupBy)
import qualified Data.Vector as V
import qualified Data.Vector.Unboxed.Mutable as M

type Tok = (Bool, (Int, Int))
type Diff= [(Op,Tok)]
type Hunk = ((Int, Int), [(Op, Tok)])

type BS = B.ByteString
type TV = V.Vector Tok

data Op
  = Remove
  | Keep
  | Add
  deriving (Show, Eq)

data DiffEnv =
  DiffEnv
    { deD1 :: BS
    , deD2 :: BS
    , deT1 :: TV
    , deT2 :: TV
    , deS :: Int
    , deE :: Int
    , deL :: Int
    , deW :: Int
    , deA :: Int
    , deB :: Int
    , deVS :: V.Vector Int
    , deVE :: V.Vector Int
    , deTrans :: Bool
    }
  deriving (Show)

substr b e = B.take (e - b) . B.drop b

toksMatch x y DiffEnv {deT1 = t1, deT2 = t2, deD1 = d1, deD2 = d2} =
  toksMatch' x y d1 d2 t1 t2

toksMatch' x y d1 d2 t1 t2 =
  let (isTok1, (b1, e1)) = t1 V.! x
      (isTok2, (b2, e2)) = t2 V.! y
   in isTok1 == isTok2 && substr b1 e1 d1 == substr b2 e2 d2

stripEqToks :: BS -> BS -> TV -> TV -> (Diff, Diff, TV, TV)
stripEqToks d1 d2 t1 t2 = (pre, post, t1', t2')
  where
    l1 = V.length t1
    l2 = V.length t2
    firstDiff i
      | i < l1 && i < l2 && toksMatch' i i d1 d2 t1 t2 = firstDiff (i + 1)
      | otherwise = i
    b = firstDiff 0
    lastDiff i
      | l1 - i - 1 >= b &&
          l2 - i - 1 >= b && toksMatch' (l1 - i - 1) (l2 - i - 1) d1 d2 t1 t2 =
        lastDiff (i + 1)
      | otherwise = i
    e = lastDiff 0
    pre = map (\t -> (Keep,t)) . V.toList . V.take b $ t1
    post = map (\t -> (Keep,t)) . V.toList . V.drop (l1-e) $ t1
    t1' = V.slice b (l1-e-b) t1
    t2' = V.slice b (l2-e-b) t2

diffToks :: BS -> BS -> TV -> TV -> Diff
diffToks d1 d2 t1' t2' = pre ++ res ++ post
  where
    (pre, post, t1, t2) = stripEqToks d1 d2 t1' t2'
    res
      | V.null t1 = map (\t -> (Add,t)) (V.toList t2)
      | V.null t2 = map (\t -> (Remove,t)) (V.toList t1)
      | V.length t1 >= V.length t2 =
        diffToks' $
        DiffEnv
          { deD1 = d1
          , deD2 = d2
          , deT1 = t1
          , deT2 = t2
          , deS = 0
          , deE = V.length t1
          , deL = V.length t1
          , deW = V.length t2
          , deA = 0
          , deB = V.length t2
          , deVS = V.fromList [0 .. V.length t2]
          , deVE = V.fromList $ reverse [0 .. V.length t2]
          , deTrans = False
          }
      | otherwise =
        diffToks' $
        DiffEnv
          { deD1 = d2
          , deD2 = d1
          , deT1 = t2
          , deT2 = t1
          , deS = 0
          , deE = V.length t2
          , deL = V.length t2
          , deW = V.length t1
          , deA = 0
          , deB = V.length t1
          , deVS = V.fromList [0 .. V.length t1]
          , deVE = V.fromList $ reverse [0 .. V.length t1]
          , deTrans = True
          }

minIndexFwd =
  V.minIndexBy
    (\x y ->
       if x <= y
         then LT
         else GT --basically normal V.minIndex
     )

minIndexRev =
  V.minIndexBy
    (\x y ->
       if x < y
         then LT
         else GT --picks the last minimum
     )

diffToks' :: DiffEnv -> Diff
diffToks' de@DiffEnv {deS = s, deE = e, deA = a, deB = b, deTrans = trans} =
  diff
  where
    mid = (s + e) `div` 2
    vecSmid = vecS mid
    vecEmid = vecE mid
    vecS = vec -- "forward" operation
      where
        vec i
          | i == s = deVS de
          | i > s = V.fromList . upd i . vec $ pred i
          | otherwise = error "Internal bounds check failure"
        upd i v = i : go 1 i
          where
            go j up
              | j > deW de = []
              | otherwise =
                let left = v V.! j
                    upleft = v V.! pred j
                    keep
                      | toksMatch (pred i) (pred j) de = min upleft
                      | otherwise = id
                    res = keep $ min (succ up) (succ left)
                 in res : go (succ j) res
    vecE = vec -- "backward" operation
      where
        vec i
          | i == e = deVE de
          | i < e = V.fromList . reverse . upd i . vec $ succ i
          | otherwise = error "Internal bounds check failure"
        upd i v = (deL de - i) : go (pred $ deW de) (deL de - i)
          where
            go j down
              | j < 0 = []
              | otherwise =
                let right = v V.! j
                    downright = v V.! succ j
                    keep
                      | toksMatch i j de = min downright
                      | otherwise = id
                    res = keep $ min (succ down) (succ right)
                 in res : go (pred j) res
    {- Now, find the optimal point for splitting.
     -
     - Heuristics A: if 2 paths are completely same, prefer deletion first;
     - which is done by choosing the 'upper' of two possibilities
     - preferentially (or 'lower' ie 'more to the right' in case of transposed
     - matrix) -}
    opt =
      (a +) .
      (if trans
         then minIndexRev
         else minIndexFwd) $
      V.zipWith (+) (slice vecSmid) (slice vecEmid)
      where
        slice = V.slice a (succ $ b - a)
    diff
      | s > e =
        error $
        "Internal failure -- recursion off limits: " <>
        show s <> " vs " <> show e
      | s == e = map (\i -> (Add, deT2 de V.! i)) [a .. pred b]
      | succ s == e =
        let vecLS = deVS de
            vecRE = deVE de
            vecLE = vecE s
            vecRS = vecS e
            sumL = V.zipWith (+) vecLS vecLE
            sumR = V.zipWith (+) vecRS vecRE
            {- This is getting a bit complicted. In the non-transposed case, we
             - want to select one Remove/Keep surrounded by 0-n Add ops, possibly
             - from both sides. The chosen path must belong to the best paths
             - (bidirectional matrix sums must match the minimum at (s,a) and
             - (b,e) on all steps), AND at the same time the path must be
             - admissible for the edit operations (ie, it has to `backtrack
             - well`). Also, it should follow Heuristic A that says that Remove
             - and Keep ops should go earlier than Add ops (or vice versa if
             - transposed). -}
            totalCost = sumL V.! a
            sCost = vecLS V.! a
            eCost = vecRS V.! b
            doKeep
              | eCost - sCost == succ (b - a) = False
              | eCost - sCost == pred (b - a) = True
              | otherwise =
                error $
                "Internal check failure -- costs seem broken: " <>
                show [sCost, eCost, a, b]
            jumpPos =
              (if trans {- Heuristic A applies here -}
                 then last
                 else head)
                [ i
                | i <-
                    [a .. if doKeep
                            then pred b
                            else b]
                , vecLS V.! i == sCost - a + i
                , sumL V.! i == totalCost
                , if doKeep
                    then vecLS V.! i == vecRS V.! succ i
                    else succ (vecLS V.! i) == vecRS V.! i
                , if doKeep
                    then sumR V.! succ i == totalCost
                    else sumR V.! i == totalCost
                , not doKeep || toksMatch s i de
                ]
            jumpEnd =
              if doKeep
                then jumpPos + 1
                else jumpPos
         in map
              (\i ->
                 ( if trans
                     then Remove
                     else Add
                 , deT2 de V.! i))
              [a .. pred jumpPos] ++
            [ if trans
                then if doKeep
                       then (Keep, deT2 de V.! jumpPos)
                       else (Add, deT1 de V.! s)
                else ( if doKeep
                         then Keep
                         else Remove
                     , deT1 de V.! s)
            ] ++
            map
              (\i ->
                 ( if trans
                     then Remove
                     else Add
                 , deT2 de V.! i))
              [jumpEnd .. pred b]
      | otherwise =
        diffToks' de {deE = mid, deVE = vecEmid, deB = opt} ++
        diffToks' de {deS = mid, deVS = vecSmid, deA = opt}

pprDiff :: BS -> BS -> Diff -> [BS]
pprDiff d1 d2 = map (pprDiff1 d1 d2)

pprDiff1 d1 d2 (op, (tok, (i, j))) =
  fromString pfx <> escNewlines (substr i j s)
  where
    pfx =
      case (op, tok) of
        (Add, True) -> "+|"
        (Remove, True) -> "-|"
        (Keep, True) -> " |"
        (Add, False) -> "+."
        (Remove, False) -> "-."
        (Keep, False) -> " ."
    s =
      case op of
        Add -> d2
        _ -> d1

escNewlines = BL.toStrict . BB.toLazyByteString . escNewlines'

escNewlines' s
  | B.null s = mempty
  | B.head s == BI.c2w '\n' = BB.stringUtf8 "\\n" <> escNewlines' (B.tail s)
  | B.head s == BI.c2w '\\' = BB.stringUtf8 "\\\\" <> escNewlines' (B.tail s)
  | otherwise = BB.word8 (B.head s) <> escNewlines' (B.tail s)

hunks :: Int -> Diff -> [Hunk]
hunks ctxt = go 0 0 0 0 [] . groupBy ((==) `on` fst)
  where
    go _ _ bi bj backlog [] =
      if null backlog
        then []
        else [((bi, bj), backlog)]
    go i j bi bj backlog (g:gs) =
      case fst (head g) of
        Remove -> go (i + length g) j bi bj (backlog ++ g) gs
        Add -> go i (j + length g) bi bj (backlog ++ g) gs
        Keep ->
          let lg = length g
           in if lg <=
                 (if null backlog || null g
                    then ctxt
                    else 2 * ctxt)
                then go (i + lg) (j + lg) bi bj (backlog ++ g) gs
                else (if null backlog
                        then id
                        else (:) ((bi, bj), backlog ++ take ctxt g)) $
                     go
                       (i + lg)
                       (j + lg)
                       (i + lg - ctxt)
                       (j + lg - ctxt)
                       (if null gs
                          then []
                          else drop (lg - ctxt) g)
                       gs

hunkHdr i j = fromString $ "@@ -" ++ show i ++ " +" ++ show j ++ " @@"

lineSep = fromString "\n"

pprHunk :: BS -> BS -> Hunk -> BS
pprHunk d1 d2 ((i, j), toks) =
  B.intercalate lineSep (hunkHdr i j : pprDiff d1 d2 toks)

pprHunks d1 d2 = B.intercalate lineSep . map (pprHunk d1 d2)
